v3.8.0 update (#2082)
* 3.8 update * fix Markus' name --------- Co-authored-by: yuzhai <yuzhai@nvidia.com>
This commit is contained in:
@@ -73,7 +73,7 @@ public:
|
||||
using ThreadEpilogueOp = typename CollectiveEpilogue::ThreadEpilogueOp;
|
||||
|
||||
using Sm100BlkScaledConfig = typename CollectiveMainloop::Sm100BlkScaledConfig;
|
||||
|
||||
|
||||
static constexpr bool epilogue_scalefactor_generation = not cute::is_same_v<typename ThreadEpilogueOp::ElementBlockScaleFactor, void>;
|
||||
static constexpr int32_t SFD_VectorSize = epilogue_scalefactor_generation ? ThreadEpilogueOp::SFVecSize : SFVecSize;
|
||||
using ElementSFD = cute::conditional_t<epilogue_scalefactor_generation, typename ThreadEpilogueOp::ElementBlockScaleFactor, void>;
|
||||
|
||||
@@ -1201,25 +1201,30 @@ public:
|
||||
GemmOperationBase<Operator_>(name) {
|
||||
|
||||
this->description_.gemm_kind = GemmKind::kGrouped;
|
||||
this->description_.kind = OperationKind::kGroupedGemm;
|
||||
this->threadblock_count = Operator::sufficient();
|
||||
}
|
||||
|
||||
private:
|
||||
int threadblock_count;
|
||||
|
||||
protected:
|
||||
|
||||
/// Constructs the arguments structure given the configuration and arguments
|
||||
static Status construct_arguments_(
|
||||
Status construct_arguments_(
|
||||
OperatorArguments &op_args,
|
||||
GemmGroupedConfiguration const *config) {
|
||||
GemmGroupedConfiguration const *config) const {
|
||||
|
||||
op_args.problem_count = config->problem_count;
|
||||
op_args.threadblock_count = config->threadblock_count;
|
||||
op_args.threadblock_count = threadblock_count;
|
||||
|
||||
return Status::kSuccess;
|
||||
}
|
||||
|
||||
/// Constructs the arguments structure given the configuration and arguments
|
||||
static Status update_arguments_(
|
||||
Status update_arguments_(
|
||||
OperatorArguments &op_args,
|
||||
GemmGroupedArguments const *arguments) {
|
||||
GemmGroupedArguments const *arguments) const {
|
||||
|
||||
if (arguments->pointer_mode == ScalarPointerMode::kHost) {
|
||||
|
||||
@@ -1243,6 +1248,8 @@ protected:
|
||||
return Status::kErrorInvalidProblem;
|
||||
}
|
||||
|
||||
op_args.threadblock_count = threadblock_count;
|
||||
op_args.problem_count = arguments->problem_count;
|
||||
op_args.problem_sizes = arguments->problem_sizes;
|
||||
|
||||
op_args.ptr_A = static_cast<ElementA **>(arguments->ptr_A);
|
||||
|
||||
@@ -36,9 +36,17 @@
|
||||
|
||||
#include "cutlass/cutlass.h"
|
||||
#include "cutlass/detail/collective.hpp"
|
||||
#include "cutlass/array.h"
|
||||
#include "cutlass/array_subbyte.h"
|
||||
#include "cutlass/library/library.h"
|
||||
#include "library_internal.h"
|
||||
#include "cutlass/gemm/dispatch_policy.hpp"
|
||||
#include "cutlass/util/packed_stride.hpp"
|
||||
#include "cutlass/util/mixed_dtype_utils.hpp"
|
||||
#include "cutlass/util/device_memory.h"
|
||||
#include "cutlass/util/reference/device/tensor_fill.h"
|
||||
#include "cutlass/util/reference/device/tensor_compare.h"
|
||||
#include "cute/tensor.hpp"
|
||||
#include <unordered_map>
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
@@ -65,7 +73,7 @@ public:
|
||||
using ElementAccumulator = typename Operator::ElementAccumulator;
|
||||
using ElementCompute = typename Operator::EpilogueOutputOp::ElementCompute;
|
||||
|
||||
private:
|
||||
protected:
|
||||
GemmDescription description_;
|
||||
|
||||
public:
|
||||
@@ -178,7 +186,23 @@ public:
|
||||
|
||||
/// Constructor
|
||||
GemmUniversal3xOperation(char const *name = "unknown_gemm"):
|
||||
GemmOperation3xBase<Operator_>(name, GemmKind::kUniversal) {}
|
||||
GemmOperation3xBase<Operator_>(name, GemmKind::kUniversal) {
|
||||
if constexpr (Operator::ArchTag::kMinComputeCapability == 90) {
|
||||
dim3 cluster_dims(
|
||||
cute::size<0>(typename Operator::GemmKernel::ClusterShape{}),
|
||||
cute::size<1>(typename Operator::GemmKernel::ClusterShape{}),
|
||||
cute::size<2>(typename Operator::GemmKernel::ClusterShape{}));
|
||||
uint32_t threads_per_block = Operator::GemmKernel::MaxThreadsPerBlock;
|
||||
void const* kernel_ptr = (void*)(device_kernel<typename Operator::GemmKernel>);
|
||||
max_active_clusters = cutlass::KernelHardwareInfo::query_device_max_active_clusters(
|
||||
cluster_dims,
|
||||
threads_per_block,
|
||||
kernel_ptr);
|
||||
}
|
||||
}
|
||||
|
||||
private:
|
||||
int max_active_clusters{};
|
||||
|
||||
protected:
|
||||
|
||||
@@ -227,10 +251,119 @@ protected:
|
||||
}
|
||||
};
|
||||
|
||||
/// Constructs the arguments structure given the configuration and arguments
|
||||
static Status update_arguments_(
|
||||
template<template<int, class, class> class Policy, int Stages, class ClusterShape, class KernelSchedule>
|
||||
static constexpr bool is_mixed_dtype_mainloop_(Policy<Stages, ClusterShape, KernelSchedule> policy) {
|
||||
return (cute::is_same_v<Policy<Stages, ClusterShape, KernelSchedule>,
|
||||
cutlass::gemm::MainloopSm90TmaGmmaRmemAWarpSpecializedMixedInput<Stages, ClusterShape, KernelSchedule>>);
|
||||
}
|
||||
|
||||
template <class DispatchPolicy>
|
||||
static constexpr bool is_mixed_dtype_mainloop_(DispatchPolicy) {
|
||||
return false;
|
||||
}
|
||||
|
||||
template <
|
||||
typename ElementWide,
|
||||
typename ElementNarrow,
|
||||
typename ElementScaleMainloop,
|
||||
class ActualStrideAB,
|
||||
Sm90MixedInputWiderOperand wider_operand,
|
||||
bool is_n4w8,
|
||||
typename ElementScale,
|
||||
typename ElementZero,
|
||||
class Layout_SZ>
|
||||
static void dequantize_encode_(
|
||||
OperatorArguments &operator_args,
|
||||
GemmUniversalArguments const *arguments) {
|
||||
GemmUniversalArguments const *arguments,
|
||||
cudaStream_t stream,
|
||||
const int &problem_mn,
|
||||
const int &problem_k,
|
||||
const int &options_l,
|
||||
const int &options_g,
|
||||
ElementScale *ptr_S,
|
||||
ElementZero *ptr_Z,
|
||||
const size_t &SZ_size,
|
||||
Layout_SZ layout_SZ
|
||||
) {
|
||||
|
||||
auto shape_AB = cute::make_shape(problem_mn, problem_k, options_l);
|
||||
auto stride_AB = cutlass::make_cute_packed_stride(ActualStrideAB{}, shape_AB);
|
||||
auto layout_AB = cute::make_layout(shape_AB, stride_AB);
|
||||
auto *ptr_dequantized_AB = static_cast<ElementWide *>(arguments->dequantized_AB);
|
||||
const ElementNarrow *ptr_AB = nullptr;
|
||||
if constexpr(wider_operand == Sm90MixedInputWiderOperand::A) {
|
||||
ptr_AB = static_cast<const ElementNarrow *>(arguments->B);
|
||||
}
|
||||
else {
|
||||
ptr_AB = static_cast<const ElementNarrow *>(arguments->A);
|
||||
}
|
||||
dequantize(ptr_dequantized_AB, ptr_AB, layout_AB, ptr_S, ptr_Z, layout_SZ, options_g, stream);
|
||||
if constexpr(is_n4w8) {
|
||||
size_t AB_size = cute::size(layout_AB);
|
||||
cutlass::int4b_t *encoded_AB = static_cast<cutlass::int4b_t *>(arguments->encoded_AB);
|
||||
unified_encode_int4b(ptr_AB, encoded_AB, AB_size);
|
||||
if constexpr(wider_operand == Sm90MixedInputWiderOperand::A) {
|
||||
operator_args.mainloop.ptr_B = static_cast<ElementNarrow const *>(encoded_AB);
|
||||
}
|
||||
else {
|
||||
operator_args.mainloop.ptr_A = static_cast<ElementNarrow const *>(encoded_AB);
|
||||
}
|
||||
ElementScaleMainloop *ptr_packed_Scale = static_cast<ElementScaleMainloop *>(arguments->packed_Scale);
|
||||
pack_scale_fp8(ptr_S, ptr_packed_Scale, SZ_size);
|
||||
}
|
||||
}
|
||||
|
||||
template <
|
||||
typename ElementAB,
|
||||
class ActualStrideAB,
|
||||
class LayoutAB_Reordered,
|
||||
class LayoutAtomQuant,
|
||||
Sm90MixedInputWiderOperand wider_operand>
|
||||
static void handle_shuffle_tensor_(
|
||||
OperatorArguments &operator_args,
|
||||
GemmUniversalArguments const *arguments,
|
||||
const int &problem_mn,
|
||||
const int &problem_k,
|
||||
const int &options_l) {
|
||||
|
||||
auto shape_AB = cute::make_shape(problem_mn, problem_k, options_l);
|
||||
auto stride_AB = cutlass::make_cute_packed_stride(ActualStrideAB{}, shape_AB);
|
||||
auto layout_AB = cute::make_layout(shape_AB, stride_AB);
|
||||
LayoutAB_Reordered layout_AB_reordered = cute::tile_to_shape(LayoutAtomQuant{}, shape_AB);
|
||||
if constexpr(wider_operand == Sm90MixedInputWiderOperand::A) {
|
||||
operator_args.mainloop.dB = layout_AB_reordered;
|
||||
}
|
||||
else {
|
||||
operator_args.mainloop.dA = layout_AB_reordered;
|
||||
}
|
||||
if (arguments->generate_dequantized_AB) {
|
||||
size_t AB_size = cute::size(layout_AB);
|
||||
ElementAB *AB_reordered = cutlass::device_memory::allocate<ElementAB>(AB_size);
|
||||
const ElementAB *AB_src = nullptr;
|
||||
if constexpr(wider_operand == Sm90MixedInputWiderOperand::A) {
|
||||
AB_src = static_cast<const ElementAB *>(operator_args.mainloop.ptr_B);
|
||||
}
|
||||
else {
|
||||
AB_src = static_cast<const ElementAB *>(operator_args.mainloop.ptr_A);
|
||||
}
|
||||
reorder_tensor(AB_src, layout_AB, AB_reordered, layout_AB_reordered);
|
||||
ElementAB *AB_dst = static_cast<ElementAB *>(arguments->encoded_AB);
|
||||
cutlass::device_memory::copy_device_to_device(AB_dst, AB_reordered, AB_size);
|
||||
cutlass::device_memory::free(AB_reordered);
|
||||
if constexpr(wider_operand == Sm90MixedInputWiderOperand::A) {
|
||||
operator_args.mainloop.ptr_B = AB_dst;
|
||||
}
|
||||
else {
|
||||
operator_args.mainloop.ptr_A = AB_dst;
|
||||
}
|
||||
}
|
||||
}
|
||||
|
||||
/// Constructs the arguments structure given the configuration and arguments
|
||||
Status update_arguments_(
|
||||
OperatorArguments& operator_args,
|
||||
GemmUniversalArguments const* arguments,
|
||||
cudaStream_t stream = nullptr) const {
|
||||
Status status = Status::kSuccess;
|
||||
|
||||
status = UpdateFusionArgs<decltype(operator_args.epilogue.thread)>::update_(
|
||||
@@ -286,24 +419,173 @@ protected:
|
||||
operator_args.epilogue.ptr_C = static_cast<ElementC const *>(arguments->C);
|
||||
operator_args.epilogue.ptr_D = static_cast<ElementD *>(arguments->D);
|
||||
|
||||
operator_args.mainloop.dA = cute::make_int_tuple_from<typename Operator::GemmKernel::StrideA>(
|
||||
// Stride{A,B} is a Layout if and only if:
|
||||
// (1) This is a mixed dtype kernel, and
|
||||
// (2) This mixed dtype kernel is using shuffling, and
|
||||
// (3) sizeof(narrow_type) == 4 or 8 bits, and
|
||||
// (4) sizeof(wide_type) == 16 bits.
|
||||
// If A/B has the narrow data type, Stride{A/B} will be a Layout
|
||||
constexpr bool is_StrideA_Layout = cute::is_layout<typename CollectiveMainloop::StrideA>::value;
|
||||
constexpr bool is_StrideB_Layout = cute::is_layout<typename CollectiveMainloop::StrideB>::value;
|
||||
static_assert(!(is_StrideA_Layout && is_StrideB_Layout), "Incorrect kernel configuration: StrideA and StrideB are both cute::Layout");
|
||||
if constexpr(!is_StrideA_Layout) {
|
||||
operator_args.mainloop.dA = cute::make_int_tuple_from<typename Operator::GemmKernel::StrideA>(
|
||||
arguments->lda, arguments->batch_stride_A);
|
||||
operator_args.mainloop.dB = cute::make_int_tuple_from<typename Operator::GemmKernel::StrideB>(
|
||||
}
|
||||
if constexpr(!is_StrideB_Layout) {
|
||||
operator_args.mainloop.dB = cute::make_int_tuple_from<typename Operator::GemmKernel::StrideB>(
|
||||
arguments->ldb, arguments->batch_stride_B);
|
||||
}
|
||||
operator_args.epilogue.dC = cute::make_int_tuple_from<typename Operator::GemmKernel::StrideC>(
|
||||
arguments->ldc, arguments->batch_stride_C);
|
||||
operator_args.epilogue.dD = operator_args.epilogue.dC;
|
||||
|
||||
using MainloopPolicy = typename CollectiveMainloop::DispatchPolicy;
|
||||
if constexpr(is_mixed_dtype_mainloop_(MainloopPolicy{})) {
|
||||
int problem_m = arguments->problem_size.m();
|
||||
int problem_n = arguments->problem_size.n();
|
||||
int problem_k = arguments->problem_size.k();
|
||||
int options_l = arguments->batch_count;
|
||||
|
||||
constexpr Sm90MixedInputWiderOperand wider_operand =
|
||||
(cutlass::sizeof_bits<ElementA>::value > cutlass::sizeof_bits<ElementB>::value) ?
|
||||
Sm90MixedInputWiderOperand::A : Sm90MixedInputWiderOperand::B;
|
||||
using ElementWide = std::conditional_t<wider_operand == Sm90MixedInputWiderOperand::A, ElementA, ElementB>;
|
||||
using ElementNarrow = std::conditional_t<wider_operand == Sm90MixedInputWiderOperand::A, ElementB, ElementA>;
|
||||
|
||||
constexpr bool has_scale = !std::is_same_v<typename CollectiveMainloop::ElementScale, void>;
|
||||
constexpr bool has_zero = !std::is_same_v<typename CollectiveMainloop::ElementZero, void>;
|
||||
if constexpr(has_scale) {
|
||||
int options_g = problem_k;
|
||||
int scale_k = (problem_k + options_g - 1) / options_g;
|
||||
|
||||
constexpr bool is_A4B8 = (
|
||||
cutlass::is_same_v<ElementA, cutlass::int4b_t> &&
|
||||
(cutlass::is_same_v<ElementB, cutlass::float_e4m3_t> ||
|
||||
cutlass::is_same_v<ElementB, cutlass::float_e5m2_t>));
|
||||
constexpr bool is_A8B4 = (
|
||||
cutlass::is_same_v<ElementB, cutlass::int4b_t> &&
|
||||
(cutlass::is_same_v<ElementA, cutlass::float_e4m3_t> ||
|
||||
cutlass::is_same_v<ElementA, cutlass::float_e5m2_t>));
|
||||
constexpr bool is_int4_x_fp8 = is_A4B8 || is_A8B4;
|
||||
|
||||
// In int4 * fp8, ElementScale is a cutlass::Array, need to take out it's real element
|
||||
using ElementScaleMainloop = typename CollectiveMainloop::ElementScale;
|
||||
using ElementScale = typename UnderlyingElement<typename CollectiveMainloop::ElementScale>::type;
|
||||
using StrideS = typename CollectiveMainloop::StrideScale;
|
||||
// In ScaleOnly mode, we have allocated the same size of memory for arguments->Z and arguments->S
|
||||
using ElementZero = std::conditional_t<
|
||||
has_zero,
|
||||
typename CollectiveMainloop::ElementZero,
|
||||
ElementScale
|
||||
>;
|
||||
const int SZ_1st_dim = (wider_operand == Sm90MixedInputWiderOperand::A) ? problem_n : problem_m;
|
||||
const size_t SZ_size = static_cast<size_t>(SZ_1st_dim * scale_k * options_l);
|
||||
auto shape_SZ = cute::make_shape(SZ_1st_dim, scale_k, options_l);
|
||||
ElementScale *ptr_S = static_cast<ElementScale *>(arguments->Scale);
|
||||
ElementZero *ptr_Z = static_cast<ElementZero *>(arguments->Zero);
|
||||
|
||||
// 1. If arguments is initialized in profiler, S and Z needs to be allocated and filled
|
||||
if (arguments->generate_scale_and_zero) {
|
||||
// Need to fix max_dequant_val and min_dequant_val?
|
||||
const float elt_max_f = float(cutlass::platform::numeric_limits<ElementScale>::max());
|
||||
const float max_dequant_val = elt_max_f * 0.25f;
|
||||
const float min_dequant_val = 0.5f;
|
||||
const float scale_max = max_dequant_val / elt_max_f;
|
||||
const float scale_min = min_dequant_val / elt_max_f;
|
||||
uint64_t seed = 2023;
|
||||
cutlass::reference::device::BlockFillRandomUniform(
|
||||
ptr_S, SZ_size, seed, ElementScale(scale_max), ElementScale(scale_min));
|
||||
|
||||
// In ScaleOnly mode, set Z as zero for generating dequantized A or B
|
||||
const float zero_max = has_zero ? 2.0f : 0.0f;
|
||||
const float zero_min = has_zero ? -2.0f : 0.0f;
|
||||
cutlass::reference::device::BlockFillRandomUniform(
|
||||
ptr_Z, SZ_size, seed, ElementZero(zero_max), ElementZero(zero_min));
|
||||
} // End of "if (arguments->generate_scale_and_zero)"
|
||||
|
||||
// 2. Generate the dequantized A or B for verification
|
||||
if (arguments->generate_dequantized_AB) {
|
||||
StrideS stride_SZ = cutlass::make_cute_packed_stride(StrideS{}, shape_SZ);
|
||||
auto layout_SZ = cute::make_layout(shape_SZ, stride_SZ);
|
||||
if constexpr(wider_operand == Sm90MixedInputWiderOperand::A) {
|
||||
if constexpr(is_StrideB_Layout) {
|
||||
// The generator only generates row-major A and col-major B at the moment
|
||||
// Need a way to read out the actual layout of B later
|
||||
using ActualLayoutB = cutlass::layout::ColumnMajor;
|
||||
using ActualStrideB = cutlass::detail::TagToStrideB_t<ActualLayoutB>;
|
||||
dequantize_encode_<ElementWide, ElementNarrow, ElementScaleMainloop, ActualStrideB, wider_operand, is_A8B4>(
|
||||
operator_args, arguments, stream, problem_m, problem_k, options_l, options_g, ptr_S, ptr_Z, SZ_size, layout_SZ);
|
||||
}
|
||||
else {
|
||||
using ActualStrideB = typename CollectiveMainloop::StrideB;
|
||||
dequantize_encode_<ElementWide, ElementNarrow, ElementScaleMainloop, ActualStrideB, wider_operand, is_A8B4>(
|
||||
operator_args, arguments, stream, problem_m, problem_k, options_l, options_g, ptr_S, ptr_Z, SZ_size, layout_SZ);
|
||||
}
|
||||
}
|
||||
else {
|
||||
if constexpr(is_StrideA_Layout) {
|
||||
// The generator only generates row-major A and col-major B at the moment
|
||||
// Need a way to read out the actual layout of A later
|
||||
using ActualLayoutA = cutlass::layout::RowMajor;
|
||||
using ActualStrideA = cutlass::detail::TagToStrideA_t<ActualLayoutA>;
|
||||
dequantize_encode_<ElementWide, ElementNarrow, ElementScaleMainloop, ActualStrideA, wider_operand, is_A4B8>(
|
||||
operator_args, arguments, stream, problem_m, problem_k, options_l, options_g, ptr_S, ptr_Z, SZ_size, layout_SZ);
|
||||
}
|
||||
else {
|
||||
using ActualStrideA = typename CollectiveMainloop::StrideA;
|
||||
dequantize_encode_<ElementWide, ElementNarrow, ElementScaleMainloop, ActualStrideA, wider_operand, is_A4B8>(
|
||||
operator_args, arguments, stream, problem_m, problem_k, options_l, options_g, ptr_S, ptr_Z, SZ_size, layout_SZ);
|
||||
}
|
||||
} // End of "if constexpr(wider_operand == Sm90MixedInputWiderOperand::A)"
|
||||
arguments->dequantized_AB_ready[0] = true;
|
||||
} // End of "if (arguments->generate_dequantized_AB)"
|
||||
|
||||
// 3. Put arguments in mainloop
|
||||
if constexpr(is_int4_x_fp8) {
|
||||
operator_args.mainloop.ptr_S = static_cast<ElementScaleMainloop const*>(arguments->packed_Scale);
|
||||
}
|
||||
else {
|
||||
operator_args.mainloop.ptr_S = static_cast<ElementScale const*>(arguments->Scale);
|
||||
}
|
||||
operator_args.mainloop.dS = cutlass::make_cute_packed_stride(StrideS{}, shape_SZ);
|
||||
operator_args.mainloop.group_size = options_g;
|
||||
if constexpr(has_zero) {
|
||||
operator_args.mainloop.ptr_Z = static_cast<ElementZero const*>(arguments->Zero);
|
||||
}
|
||||
} // End of "if constexpr(has_scale)"
|
||||
|
||||
// Handle the shuffling
|
||||
using ValueShuffle = std::conditional_t<
|
||||
cutlass::sizeof_bits<ElementNarrow>::value == 4,
|
||||
cute::Layout<cute::Shape<cute::_2,cute::_4>, cute::Stride<cute::_4,cute::_1>>,
|
||||
cute::Layout<cute::Shape<cute::_2,cute::_2>, cute::Stride<cute::_2,cute::_1>>
|
||||
>;
|
||||
constexpr int NumShuffleAtoms = 1;
|
||||
using MmaAtomShape = cute::Layout<cute::Shape<cute::_1,cute::Int<NumShuffleAtoms>>>;
|
||||
using LayoutAtomQuant = decltype(compute_memory_reordering_atom<ElementWide, MmaAtomShape, ValueShuffle>());
|
||||
// The generator only generates row-major A and col-major B at the moment
|
||||
// Need a way to read out the actual layout and stride of A/B later
|
||||
if constexpr(wider_operand == Sm90MixedInputWiderOperand::A && is_StrideB_Layout) {
|
||||
using ActualLayoutB = cutlass::layout::ColumnMajor;
|
||||
using ActualStrideB = cutlass::detail::TagToStrideB_t<ActualLayoutB>;
|
||||
using LayoutB_Reordered = typename CollectiveMainloop::StrideB;
|
||||
handle_shuffle_tensor_<ElementB, ActualStrideB, LayoutB_Reordered, LayoutAtomQuant, wider_operand>(
|
||||
operator_args, arguments, problem_n, problem_k, options_l);
|
||||
}
|
||||
if constexpr(wider_operand == Sm90MixedInputWiderOperand::B && is_StrideA_Layout) {
|
||||
using ActualLayoutA = cutlass::layout::RowMajor;
|
||||
using ActualStrideA = cutlass::detail::TagToStrideA_t<ActualLayoutA>;
|
||||
using LayoutA_Reordered = typename CollectiveMainloop::StrideA;
|
||||
handle_shuffle_tensor_<ElementA, ActualStrideA, LayoutA_Reordered, LayoutAtomQuant, wider_operand>(
|
||||
operator_args, arguments, problem_m, problem_k, options_l);
|
||||
}
|
||||
} // End of "if constexpr(is_mixed_dtype_mainloop_(MainloopPolicy{}))"
|
||||
|
||||
/* Query device SM count and max active clusters to pass onto the kernel as an argument, where needed */
|
||||
operator_args.hw_info.sm_count = arguments->sm_count;
|
||||
if constexpr (Operator::ArchTag::kMinComputeCapability == 90) {
|
||||
dim3 cluster_dims(cute::size<0>(typename Operator::GemmKernel::ClusterShape{}),
|
||||
cute::size<1>(typename Operator::GemmKernel::ClusterShape{}),
|
||||
cute::size<2>(typename Operator::GemmKernel::ClusterShape{}));
|
||||
uint32_t threads_per_block = Operator::GemmKernel::MaxThreadsPerBlock;
|
||||
void const* kernel_ptr = (void*)(device_kernel<typename Operator::GemmKernel>);
|
||||
operator_args.hw_info.max_active_clusters = cutlass::KernelHardwareInfo::query_device_max_active_clusters(
|
||||
cluster_dims, threads_per_block, kernel_ptr);
|
||||
operator_args.hw_info.max_active_clusters = max_active_clusters;
|
||||
}
|
||||
if constexpr (!std::is_const_v<decltype(operator_args.scheduler.max_swizzle_size)>) {
|
||||
operator_args.scheduler.max_swizzle_size = arguments->swizzle_size;
|
||||
@@ -356,7 +638,10 @@ public:
|
||||
return status;
|
||||
}
|
||||
|
||||
return Operator::can_implement(args);
|
||||
Status can_impl = Operator::can_implement(args);
|
||||
|
||||
//return Operator::can_implement(args);
|
||||
return can_impl;
|
||||
}
|
||||
|
||||
/// Gets the host-side workspace
|
||||
@@ -397,7 +682,7 @@ public:
|
||||
cudaStream_t stream = nullptr) const override {
|
||||
|
||||
OperatorArguments args;
|
||||
Status status = update_arguments_(args, static_cast<GemmUniversalArguments const *>(arguments_ptr));
|
||||
Status status = update_arguments_(args, static_cast<GemmUniversalArguments const *>(arguments_ptr), stream);
|
||||
if (status != Status::kSuccess) {
|
||||
return status;
|
||||
}
|
||||
|
||||
@@ -0,0 +1,330 @@
|
||||
/***************************************************************************************************
|
||||
* Copyright (c) 2025 - 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
|
||||
* SPDX-License-Identifier: BSD-3-Clause
|
||||
*
|
||||
* Redistribution and use in source and binary forms, with or without
|
||||
* modification, are permitted provided that the following conditions are met:
|
||||
*
|
||||
* 1. Redistributions of source code must retain the above copyright notice, this
|
||||
* list of conditions and the following disclaimer.
|
||||
*
|
||||
* 2. Redistributions in binary form must reproduce the above copyright notice,
|
||||
* this list of conditions and the following disclaimer in the documentation
|
||||
* and/or other materials provided with the distribution.
|
||||
*
|
||||
* 3. Neither the name of the copyright holder nor the names of its
|
||||
* contributors may be used to endorse or promote products derived from
|
||||
* this software without specific prior written permission.
|
||||
*
|
||||
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
|
||||
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
|
||||
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
|
||||
* DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
|
||||
* FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
|
||||
* DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
|
||||
* SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
|
||||
* CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
|
||||
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
|
||||
* OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
|
||||
*
|
||||
**************************************************************************************************/
|
||||
/* \file
|
||||
\brief Defines operations for all grouped GEMM operations in CUTLASS Library.
|
||||
*/
|
||||
|
||||
#pragma once
|
||||
|
||||
#include "cutlass/cutlass.h"
|
||||
#include "cutlass/detail/collective.hpp"
|
||||
#include "cutlass/gemm/dispatch_policy.hpp"
|
||||
#include "cutlass/library/library.h"
|
||||
#include "cutlass/library/util.h"
|
||||
#include "gemm_operation_3x.hpp"
|
||||
#include "library_internal.h"
|
||||
#include <unordered_map>
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
namespace cutlass::library {
|
||||
|
||||
/// **** CAUTION ****
|
||||
/// Unlike other operations, initialize() must be called when
|
||||
/// certain arguments change. See initialize() for details.
|
||||
template <typename Operator_>
|
||||
class GroupedGemmUniversal3xOperation : public GemmOperation3xBase<Operator_> {
|
||||
public:
|
||||
using Operator = Operator_;
|
||||
using OperatorArguments = typename Operator::Arguments;
|
||||
using ElementA = typename Operator::ElementA;
|
||||
using LayoutA = typename Operator::LayoutA;
|
||||
using ElementB = typename Operator::ElementB;
|
||||
using LayoutB = typename Operator::LayoutB;
|
||||
using ElementC = typename Operator::ElementC;
|
||||
using LayoutC = typename Operator::LayoutC;
|
||||
using ElementD = typename Operator::ElementD;
|
||||
using LayoutD = typename Operator::LayoutD;
|
||||
using ElementAccumulator = typename Operator::ElementAccumulator;
|
||||
using ElementCompute = typename Operator::EpilogueOutputOp::ElementCompute;
|
||||
|
||||
using CollectiveMainloop = typename Operator::CollectiveMainloop;
|
||||
using CollectiveEpilogue = typename Operator::CollectiveEpilogue;
|
||||
using ThreadEpilogueOp = typename CollectiveEpilogue::ThreadEpilogueOp;
|
||||
|
||||
private:
|
||||
mutable CudaBuffer strideA_device;
|
||||
mutable CudaBuffer strideB_device;
|
||||
mutable CudaBuffer strideC_device;
|
||||
mutable CudaBuffer strideD_device;
|
||||
mutable std::vector<typename Operator::GemmKernel::InternalStrideA> strideA_host;
|
||||
mutable std::vector<typename Operator::GemmKernel::InternalStrideB> strideB_host;
|
||||
mutable std::vector<typename Operator::GemmKernel::InternalStrideC> strideC_host;
|
||||
mutable std::vector<typename Operator::GemmKernel::InternalStrideD> strideD_host;
|
||||
|
||||
public:
|
||||
GroupedGemmUniversal3xOperation(char const* name = "unknown_gemm")
|
||||
: GemmOperation3xBase<Operator_>(name, GemmKind::kGrouped) {
|
||||
this->description_.kind = OperationKind::kGroupedGemm;
|
||||
if constexpr (Operator::ArchTag::kMinComputeCapability >= 90) {
|
||||
dim3 cluster_dims(
|
||||
cute::size<0>(typename Operator::GemmKernel::ClusterShape{}),
|
||||
cute::size<1>(typename Operator::GemmKernel::ClusterShape{}),
|
||||
cute::size<2>(typename Operator::GemmKernel::ClusterShape{}));
|
||||
uint32_t threads_per_block = Operator::GemmKernel::MaxThreadsPerBlock;
|
||||
void const* kernel_ptr = (void*)(device_kernel<typename Operator::GemmKernel>);
|
||||
max_active_clusters = cutlass::KernelHardwareInfo::query_device_max_active_clusters(
|
||||
cluster_dims,
|
||||
threads_per_block,
|
||||
kernel_ptr);
|
||||
}
|
||||
}
|
||||
|
||||
~GroupedGemmUniversal3xOperation() override = default;
|
||||
|
||||
private:
|
||||
int max_active_clusters{};
|
||||
|
||||
protected:
|
||||
template <class FusionArgs, class = void> struct UpdateFusionArgs {
|
||||
static Status update_(FusionArgs const& fusion_args, GemmGroupedArguments const& arguments) {
|
||||
// If a custom EVT is instantiated then it is the users's responsibility
|
||||
// to ensure alpha and beta are updated appropriately
|
||||
return Status::kSuccess;
|
||||
}
|
||||
};
|
||||
|
||||
template <class FusionArgs>
|
||||
struct UpdateFusionArgs<FusionArgs, cute::void_t<decltype(FusionArgs{}.alpha)>> {
|
||||
static Status update_(FusionArgs& fusion_args, GemmGroupedArguments const& arguments) {
|
||||
if (arguments.pointer_mode == ScalarPointerMode::kHost) {
|
||||
fusion_args.alpha = *static_cast<ElementCompute const*>(arguments.alpha);
|
||||
fusion_args.beta = *static_cast<ElementCompute const*>(arguments.beta);
|
||||
fusion_args.alpha_ptr = nullptr;
|
||||
fusion_args.beta_ptr = nullptr;
|
||||
fusion_args.alpha_ptr_array = nullptr;
|
||||
fusion_args.beta_ptr_array = nullptr;
|
||||
// Single alpha and beta for all groups
|
||||
fusion_args.dAlpha = {cute::_0{}, cute::_0{}, 0};
|
||||
fusion_args.dBeta = {cute::_0{}, cute::_0{}, 0};
|
||||
|
||||
return Status::kSuccess;
|
||||
}
|
||||
else if (arguments.pointer_mode == ScalarPointerMode::kDevice) {
|
||||
fusion_args.alpha = 0;
|
||||
fusion_args.beta = 0;
|
||||
fusion_args.alpha_ptr = static_cast<ElementCompute const*>(arguments.alpha);
|
||||
fusion_args.beta_ptr = static_cast<ElementCompute const*>(arguments.beta);
|
||||
return Status::kSuccess;
|
||||
}
|
||||
else {
|
||||
return Status::kErrorInvalidProblem;
|
||||
}
|
||||
}
|
||||
};
|
||||
|
||||
/// Constructs the arguments structure given the configuration and arguments
|
||||
Status
|
||||
update_arguments_(OperatorArguments& operator_args, GemmGroupedArguments const* arguments) const {
|
||||
|
||||
Status status = UpdateFusionArgs<decltype(operator_args.epilogue.thread)>::update_(
|
||||
operator_args.epilogue.thread,
|
||||
*arguments);
|
||||
if (status != Status::kSuccess) {
|
||||
return status;
|
||||
}
|
||||
|
||||
operator_args.mode = cutlass::gemm::GemmUniversalMode::kGrouped;
|
||||
operator_args.problem_shape = {
|
||||
arguments->problem_count,
|
||||
arguments->problem_sizes_3x,
|
||||
arguments->pointer_mode == ScalarPointerMode::kHost ? arguments->problem_sizes_3x_host
|
||||
: nullptr};
|
||||
operator_args.mainloop.ptr_A =
|
||||
static_cast<const typename Operator::ElementA**>(arguments->ptr_A);
|
||||
operator_args.mainloop.ptr_B =
|
||||
static_cast<const typename Operator::ElementB**>(arguments->ptr_B);
|
||||
operator_args.epilogue.ptr_C =
|
||||
static_cast<const typename Operator::ElementC**>(arguments->ptr_C);
|
||||
operator_args.epilogue.ptr_D = static_cast<typename Operator::ElementD**>(arguments->ptr_D);
|
||||
|
||||
operator_args.mainloop.dA =
|
||||
static_cast<typename Operator::GemmKernel::InternalStrideA*>(strideA_device.data());
|
||||
operator_args.mainloop.dB =
|
||||
static_cast<typename Operator::GemmKernel::InternalStrideB*>(strideB_device.data());
|
||||
operator_args.epilogue.dC =
|
||||
static_cast<typename Operator::GemmKernel::InternalStrideC*>(strideC_device.data());
|
||||
operator_args.epilogue.dD =
|
||||
static_cast<typename Operator::GemmKernel::InternalStrideD*>(strideD_device.data());
|
||||
|
||||
operator_args.hw_info.sm_count = arguments->sm_count;
|
||||
if constexpr (Operator::ArchTag::kMinComputeCapability >= 90) {
|
||||
operator_args.hw_info.max_active_clusters = max_active_clusters;
|
||||
}
|
||||
|
||||
if constexpr (Operator::ArchTag::kMinComputeCapability >= 100) {
|
||||
operator_args.hw_info.cluster_shape = dim3(
|
||||
arguments->cluster_shape.m(),
|
||||
arguments->cluster_shape.n(),
|
||||
arguments->cluster_shape.k());
|
||||
operator_args.hw_info.cluster_shape_fallback = dim3(
|
||||
arguments->cluster_shape_fallback.m(),
|
||||
arguments->cluster_shape_fallback.n(),
|
||||
arguments->cluster_shape_fallback.k());
|
||||
}
|
||||
|
||||
|
||||
return status;
|
||||
}
|
||||
|
||||
public:
|
||||
/// Returns success if the operation can proceed
|
||||
Status can_implement([[maybe_unused]] void const* configuration_ptr, void const* arguments_ptr)
|
||||
const override {
|
||||
GemmGroupedArguments const* arguments = static_cast<GemmGroupedArguments const*>(arguments_ptr);
|
||||
OperatorArguments args;
|
||||
|
||||
auto status = update_arguments_(args, arguments);
|
||||
if (status != Status::kSuccess) {
|
||||
return status;
|
||||
}
|
||||
|
||||
status = Operator::can_implement(args);
|
||||
return status;
|
||||
}
|
||||
|
||||
/// Gets the host-side workspace
|
||||
uint64_t get_host_workspace_size(void const* configuration) const override {
|
||||
return sizeof(Operator);
|
||||
}
|
||||
|
||||
/// Gets the device-side workspace
|
||||
uint64_t get_device_workspace_size(void const* configuration_ptr, void const* arguments_ptr)
|
||||
const override {
|
||||
|
||||
OperatorArguments args;
|
||||
auto status = update_arguments_(args, static_cast<GemmGroupedArguments const*>(arguments_ptr));
|
||||
if (status != Status::kSuccess) {
|
||||
return 0;
|
||||
}
|
||||
|
||||
uint64_t size = Operator::get_workspace_size(args);
|
||||
return size;
|
||||
}
|
||||
|
||||
/// Initializes the workspace
|
||||
/// **** CAUTION ****
|
||||
/// Must be called when lda, ldb, ldc, or ldd change.
|
||||
/// The CUTLASS library stores the operations in a type-
|
||||
/// erased manifest. Therefore, only this class knows
|
||||
/// the type of strideA, strideB, strideC, and strideD.
|
||||
/// Since grouped GEMM needs to allocate storage for
|
||||
/// the strides on device, the concrete type of the stride
|
||||
/// must be known in order to copy in the correct memory
|
||||
/// layout on device.
|
||||
Status initialize(
|
||||
void const* configuration_ptr,
|
||||
void* host_workspace,
|
||||
void* device_workspace,
|
||||
cudaStream_t stream = nullptr) const override {
|
||||
|
||||
auto const& config = *static_cast<GemmGroupedConfiguration const*>(configuration_ptr);
|
||||
|
||||
auto num_groups = config.problem_count;
|
||||
strideA_device =
|
||||
CudaBuffer(sizeof(typename Operator::GemmKernel::InternalStrideA) * num_groups);
|
||||
strideB_device =
|
||||
CudaBuffer(sizeof(typename Operator::GemmKernel::InternalStrideB) * num_groups);
|
||||
strideC_device =
|
||||
CudaBuffer(sizeof(typename Operator::GemmKernel::InternalStrideC) * num_groups);
|
||||
strideD_device =
|
||||
CudaBuffer(sizeof(typename Operator::GemmKernel::InternalStrideD) * num_groups);
|
||||
|
||||
strideA_host.resize(num_groups);
|
||||
strideB_host.resize(num_groups);
|
||||
strideC_host.resize(num_groups);
|
||||
strideD_host.resize(num_groups);
|
||||
for (int group_idx = 0; group_idx < num_groups; group_idx++) {
|
||||
strideA_host[group_idx] =
|
||||
cute::make_int_tuple_from<typename Operator::GemmKernel::InternalStrideA>(
|
||||
config.lda[group_idx]);
|
||||
strideB_host[group_idx] =
|
||||
cute::make_int_tuple_from<typename Operator::GemmKernel::InternalStrideB>(
|
||||
config.ldb[group_idx]);
|
||||
strideC_host[group_idx] =
|
||||
cute::make_int_tuple_from<typename Operator::GemmKernel::InternalStrideC>(
|
||||
config.ldc[group_idx]);
|
||||
strideD_host[group_idx] =
|
||||
cute::make_int_tuple_from<typename Operator::GemmKernel::InternalStrideD>(
|
||||
config.ldc[group_idx]);
|
||||
}
|
||||
CUDA_CHECK(cudaMemcpy(
|
||||
strideA_device.data(),
|
||||
strideA_host.data(),
|
||||
sizeof(typename Operator::GemmKernel::InternalStrideA) * num_groups,
|
||||
cudaMemcpyHostToDevice));
|
||||
CUDA_CHECK(cudaMemcpy(
|
||||
strideB_device.data(),
|
||||
strideB_host.data(),
|
||||
sizeof(typename Operator::GemmKernel::InternalStrideB) * num_groups,
|
||||
cudaMemcpyHostToDevice));
|
||||
CUDA_CHECK(cudaMemcpy(
|
||||
strideC_device.data(),
|
||||
strideC_host.data(),
|
||||
sizeof(typename Operator::GemmKernel::InternalStrideC) * num_groups,
|
||||
cudaMemcpyHostToDevice));
|
||||
CUDA_CHECK(cudaMemcpy(
|
||||
strideD_device.data(),
|
||||
strideD_host.data(),
|
||||
sizeof(typename Operator::GemmKernel::InternalStrideD) * num_groups,
|
||||
cudaMemcpyHostToDevice));
|
||||
|
||||
Operator* op = new (host_workspace) Operator;
|
||||
return Status::kSuccess;
|
||||
}
|
||||
|
||||
/// **** CAUTION ****
|
||||
/// initialize() must be called if lda, ldb, ldc, or ldd change.
|
||||
Status run(
|
||||
void const* arguments_ptr,
|
||||
void* host_workspace,
|
||||
void* device_workspace = nullptr,
|
||||
cudaStream_t stream = nullptr) const override {
|
||||
|
||||
OperatorArguments operator_args;
|
||||
auto const& args = *static_cast<GemmGroupedArguments const*>(arguments_ptr);
|
||||
|
||||
Status status = update_arguments_(operator_args, &args);
|
||||
if (status != Status::kSuccess) {
|
||||
return status;
|
||||
}
|
||||
|
||||
Operator* op = static_cast<Operator*>(host_workspace);
|
||||
// We need to call initialize() since we have to rebuild TMA desc for every new set of args
|
||||
status = op->run(operator_args, device_workspace, stream, nullptr, args.use_pdl);
|
||||
return status;
|
||||
}
|
||||
};
|
||||
///////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
} // namespace cutlass::library
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
@@ -64,7 +64,6 @@ void initialize_gemm_reference_operations_f8_f6_f32(Manifest &manifest);
|
||||
void initialize_block_scaled_gemm_reference_operations_fp4a_vs16(Manifest &manifest);
|
||||
void initialize_block_scaled_gemm_reference_operations_fp4a_vs32(Manifest &manifest);
|
||||
void initialize_block_scaled_gemm_reference_operations_mixed8bitsa(Manifest &manifest);
|
||||
|
||||
void initialize_gemm_reference_operations_fp8in_fp16out(Manifest &manifest);
|
||||
void initialize_gemm_reference_operations_fp8in_bf16out(Manifest &manifest);
|
||||
void initialize_gemm_reference_operations_fp8in_fp32out(Manifest &manifest);
|
||||
@@ -114,7 +113,6 @@ void initialize_reference_operations(Manifest &manifest) {
|
||||
initialize_block_scaled_gemm_reference_operations_fp4a_vs16(manifest);
|
||||
initialize_block_scaled_gemm_reference_operations_fp4a_vs32(manifest);
|
||||
initialize_block_scaled_gemm_reference_operations_mixed8bitsa(manifest);
|
||||
|
||||
}
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
@@ -37,6 +37,7 @@
|
||||
#include "cutlass/cutlass.h"
|
||||
#include "cutlass/detail/collective.hpp"
|
||||
#include "cutlass/library/library.h"
|
||||
#include "cutlass/library/util.h"
|
||||
#include "cutlass/transform/kernel/sparse_gemm_compressor.hpp" // StructuredSparseCompressor
|
||||
#include "cutlass/transform/device/transform_universal_adapter.hpp" // TransformUniversalAdapter
|
||||
#include "cutlass/util/packed_stride.hpp" // make_cute_packed_stride
|
||||
@@ -45,14 +46,6 @@
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
#define CUDA_CHECK(cuda_error) \
|
||||
{ \
|
||||
if (cuda_error != cudaSuccess) { \
|
||||
printf("cudaError %s in %s:%d\n", cudaGetErrorString(cuda_error), __func__, __LINE__ ); \
|
||||
return Status::kInvalid; \
|
||||
} \
|
||||
}
|
||||
|
||||
namespace cutlass::library {
|
||||
|
||||
///////////////////////////////////////////////////////////////////////////////////////////////////
|
||||
|
||||
@@ -330,18 +330,18 @@ static struct {
|
||||
char const *text;
|
||||
char const *pretty;
|
||||
OperationKind enumerant;
|
||||
}
|
||||
OperationKind_enumerants[] = {
|
||||
{"eq_gemm", "EqGemm", OperationKind::kEqGemm},
|
||||
} OperationKind_enumerants[] = {
|
||||
{"eq_gemm", "EqGemm", OperationKind::kEqGemm},
|
||||
{"gemm", "Gemm", OperationKind::kGemm},
|
||||
{"block_scaled_gemm", "blockScaledGemm", OperationKind::kBlockScaledGemm},
|
||||
{"rank_k", "RankK", OperationKind::kRankK},
|
||||
{"rank_2k", "Rank2K", OperationKind::kRank2K},
|
||||
{"trmm", "Trmm", OperationKind::kTrmm},
|
||||
{"symm", "Symm", OperationKind::kSymm},
|
||||
{"conv2d", "Conv2d", OperationKind::kConv2d},
|
||||
{"conv3d", "Conv3d", OperationKind::kConv3d},
|
||||
{"conv2d", "Conv2d", OperationKind::kConv2d},
|
||||
{"conv3d", "Conv3d", OperationKind::kConv3d},
|
||||
{"spgemm", "SparseGemm", OperationKind::kSparseGemm},
|
||||
{"grouped_gemm", "GroupedGemm", OperationKind::kGroupedGemm},
|
||||
};
|
||||
|
||||
/// Converts a Status enumerant to a string
|
||||
@@ -504,7 +504,6 @@ NumericTypeID_enumerants[] = {
|
||||
{"fe2m1", "FE2M1", NumericTypeID::kFE2M1},
|
||||
{"fue8m0", "FUE8M0", NumericTypeID::kFUE8M0},
|
||||
{"fue4m3", "FUE4M3", NumericTypeID::kFUE4M3},
|
||||
|
||||
{"f16", "F16", NumericTypeID::kF16},
|
||||
{"bf16", "BF16", NumericTypeID::kBF16},
|
||||
{"f32", "F32", NumericTypeID::kF32},
|
||||
@@ -577,7 +576,6 @@ int sizeof_bits(NumericTypeID type) {
|
||||
case NumericTypeID::kFE2M1: return 4;
|
||||
case NumericTypeID::kFUE8M0: return 8;
|
||||
case NumericTypeID::kFUE4M3: return 8;
|
||||
|
||||
case NumericTypeID::kF16: return 16;
|
||||
case NumericTypeID::kBF16: return 16;
|
||||
case NumericTypeID::kTF32: return 32;
|
||||
@@ -666,7 +664,6 @@ bool is_signed_type(NumericTypeID type) {
|
||||
case NumericTypeID::kFE2M1: return true;
|
||||
case NumericTypeID::kFUE8M0: return false;
|
||||
case NumericTypeID::kFUE4M3: return false;
|
||||
|
||||
case NumericTypeID::kF16: return true;
|
||||
case NumericTypeID::kBF16: return true;
|
||||
case NumericTypeID::kTF32: return true;
|
||||
@@ -707,7 +704,6 @@ bool is_float_type(NumericTypeID type) {
|
||||
case NumericTypeID::kFE2M1: return true;
|
||||
case NumericTypeID::kFUE8M0: return true;
|
||||
case NumericTypeID::kFUE4M3: return true;
|
||||
|
||||
case NumericTypeID::kF16: return true;
|
||||
case NumericTypeID::kBF16: return true;
|
||||
case NumericTypeID::kTF32: return true;
|
||||
@@ -1256,7 +1252,6 @@ bool lexical_cast(std::vector<uint8_t> &bytes, NumericTypeID type, std::string c
|
||||
*reinterpret_cast<float_e5m2_t *>(bytes.data()) = static_cast<float_e5m2_t>(tmp);
|
||||
}
|
||||
break;
|
||||
|
||||
case NumericTypeID::kFE2M3:
|
||||
{
|
||||
float tmp;
|
||||
@@ -1292,7 +1287,6 @@ bool lexical_cast(std::vector<uint8_t> &bytes, NumericTypeID type, std::string c
|
||||
*reinterpret_cast<float_ue4m3_t *>(bytes.data()) = static_cast<float_ue4m3_t>(tmp);
|
||||
}
|
||||
break;
|
||||
|
||||
case NumericTypeID::kF16:
|
||||
{
|
||||
float tmp;
|
||||
@@ -1473,7 +1467,6 @@ std::string lexical_cast(std::vector<uint8_t> &bytes, NumericTypeID type) {
|
||||
ss << tmp;
|
||||
}
|
||||
break;
|
||||
|
||||
case NumericTypeID::kF16:
|
||||
{
|
||||
float tmp = *reinterpret_cast<half_t *>(bytes.data());
|
||||
@@ -1652,7 +1645,6 @@ bool cast_from_int64(std::vector<uint8_t> &bytes, NumericTypeID type, int64_t sr
|
||||
*reinterpret_cast<float_ue4m3_t *>(bytes.data()) = static_cast<float_ue4m3_t>(float(src));
|
||||
}
|
||||
break;
|
||||
|
||||
case NumericTypeID::kF16:
|
||||
{
|
||||
*reinterpret_cast<half_t *>(bytes.data()) = static_cast<half_t>(float(src));
|
||||
@@ -1789,7 +1781,6 @@ bool cast_from_uint64(std::vector<uint8_t> &bytes, NumericTypeID type, uint64_t
|
||||
*reinterpret_cast<float_ue4m3_t *>(bytes.data()) = static_cast<float_ue4m3_t>(float(src));
|
||||
}
|
||||
break;
|
||||
|
||||
case NumericTypeID::kF16:
|
||||
{
|
||||
*reinterpret_cast<half_t *>(bytes.data()) = static_cast<half_t>(float(src));
|
||||
@@ -1927,7 +1918,6 @@ bool cast_from_double(std::vector<uint8_t> &bytes, NumericTypeID type, double sr
|
||||
*reinterpret_cast<float_ue4m3_t *>(bytes.data()) = static_cast<float_ue4m3_t>(float(src));
|
||||
}
|
||||
break;
|
||||
|
||||
case NumericTypeID::kF16:
|
||||
{
|
||||
*reinterpret_cast<half_t *>(bytes.data()) = static_cast<half_t>(float(src));
|
||||
|
||||
Reference in New Issue
Block a user